from . import account, auth, ca, cert, find, forge, ptt, relay, req, shadow, template

ENTRY_PARSERS = [
    account,
    auth,
    ca,
    cert,
    find,
    forge,
    ptt,
    relay,
    req,
    shadow,
    template,
]
